﻿/* Cane Toads Program, January 2012
 * Whitney Sanders (sanderswe@email.wofford.edu)

 * This is a parallel version of the simulation which divides up each
 * simulation among processes. This version is efficient if the size of
 * each simulations is large relative to the number of simulations.

 * DESCRIPTION:
 * To accompany an educational module written by Dr. Angela Shiflet and
 * Dr. George Shiflet, titled "Cane Toads"

 * SPECIFICATIONS:
 * The specifics for the simulation are described in the first group
 * of lines containing values for the amount a toad eats, the probability
 * a toad will hop if it's not thirsty or hungry, and so forth.
 * 
 * OUTPUT:
 * The output is printed to standard output. Output contains the number
 * of toads which are alive, dead, and migrated.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <mpi.h>

#define ALIVE 1 // toad is alive
#define AMT_DRINK 0.05 // max amount toad drinks in 1 time step 0.05
#define AMT_EAT 0.01 // max amount toad eats in 1 time step 0.05
#define AWP 1 // moisture value for water, such as an AWP
#define AWP_R1 0.4 // moisture value of cells neighboring water
#define AWP_R2 0.2 // moisture value of cell 2 cells away from water
#define AWPs 4 // number of AWPs
#define DESICCATE 0.6 // level at which desiccation occurs 0.6
#define DEAD 0 // toad is dead
#define ENERGY_HOPPING 0.002 // max amount of energy used by a hop
#define FOOD_CELL 0.03 // food value for initializing constant food grid
#define LIMITED 0.0 // moisture value for limited amount of moisture
#define MAY_EAT 0.2 // probability of eating if thirsty
#define MAY_HOP 0.5 // probability of hopping if not thirsty/hungry
#define MEAN_ENERGY 0.94 // mean energy for initialized toad
#define MEAN_FOOD 0.1 // mean value for initializing random food grid
#define MEAN_WATER 0.94 // mean water for initialized toad 0.94
#define MIGRATED 2 // toad has migrated off the grid
#define MOIST_AREA 0.5 // moisture value for moist area
#define MOISTS 3 // number of moist areas 3
#define NEED_TO_DRINK 0.8 // water level at which toad needs to drink
#define NUM_FENCED 4 // number of fenced AWPs
#define NUM_TOADS 50 // number of toads 50
#define NUM_NEIGHBORS 8 // 4, or 8 if diagonal moves are allowed
#define NUM_SIMULATIONS 100
#define NUM_STEPS 200 // number of simulation steps, 300
#define FRACTION_WATER 0.6 // fraction of prey that is water
#define PERMANENT_WATER 1.0 // moisture value for natural water
#define SIDE 20 // number of cells along one side of grid 20
#define STARVE 0.6 // level at which starvation occurs
#define STD_ENERGY 0.02 // standard deviation of toads' initial energy
#define STD_FOOD 0.02 // standard deviation of toad's initial food
#define STD_WATER 0.02 // standard deviation of toad's initial water
#define WATER_HOPPING 0.002 // max amount of water used by a hop
#define WOULD_LIKE_DRINK 0.9 // water level at which toad wants a drink
#define WOULD_LIKE_EAT 0.9 // food level at which toad wants food

// MPI macros
#define SERVER 0
#define TAG 2

// Mathematical constants
#define M_E 2.71828
#define M_PI 3.14159

int ROWLST_AWP[AWPs] = {3,11,15,9}; // array of row indices for AWPs
int COLLST_AWP[AWPs] = {5,13,3,9}; // array of column indices for AWPs
int ROWLST_MOISTAREA[MOISTS] = {13,15,3}; // array of column indices for moist areas
int COLLST_MOISTAREA[MOISTS] = {7,15,17}; // array of row indices for moist areas
int FENCED[NUM_FENCED] = {0,1,2,3}; // array of ROWLST_AWP indices of fenced AWPs

struct toad {
	int toadID; // Identification number
	double energy; // From 0.0 to 1.0 indicating energy level
	double water; // From 0.0 to 1.0 indicating amount of water
	int alive; // Toad's state; 0: dead, 1: alive, 2: migrated
	int row; // Matrix row location of the toad on grid
	int col; // Matrix column location of toad on grid
	int rowLast; // Previous row location, when looking for water
	int colLast; // Previous col location, when looking for water
	int numTimeSteps; // Number of steps in moist area
};

double food[SIDE][SIDE];
double moisture[SIDE][SIDE];
struct toad toads[NUM_TOADS];

/* ---------------------------------------------------
 * Simulation constants and globals have been defined.
 * --------------------------------------------------- */

double normalDistributionRand(double mean, double stdev, double min, double max) {
	// Called by initFood(), initToads()
	// Returns a random number generated on a normal distribution with
	// given mean and standard deviation, within the given interval

	// Algorithm: Box-Muller-Gauss Method
	// a = uniform random number in [0, 2 * pi)
	// b = stdev * sqrt( -2 * ln(c) )
	// c = uniform random number in [0, 1)
	// return r, where:
	// 	r = b * sin(a) + stdev with probability 0.5
	// 	r = b * cos(a) + stdev with probability 0.5

	// 2*pi is approximately 6.283185
	double a = (rand() % 6283185) / 1000000.0;
	double c = (rand() % 1000000) / 1000000.0;
	double b = stdev * sqrt(-2 * log(c));
	int p = rand() % 2;
	double r;
	if (p == 1) {
		r = ( b * sin(a) ) + mean;
	} else {
		r = ( b * cos(a) ) + mean;
	}
	if (r > max)
		r = max;
	if (r < min)
		r = min;
	return r;
} // normalDistributionRand()

void initFood(void) {
	// Called by init()
	// Initializes global SIDE-by-SIDE food matrix

	// Give each cell a normally distributed food value, based on
	// MEAN_FOOD and STD_FOOD, from 0.0 to 1.0
	for (int i = 0; i < SIDE; i++) {
		for (int j = 0; j < SIDE; j++) {
			// Option 1: use a normal distribution
			//double val = normalDistributionRand(MEAN_FOOD, STD_FOOD, 0.0, 1.0);
			//food[i][j] = val;

			// Option 2: Use constant food values
			food[i][j] = FOOD_CELL;
		}
	}

	// Initialize north, south, east borders at -1
	for (int i = 0; i < SIDE; i++) // North
		food[0][i] = -1;
	for (int i = 0; i < SIDE; i++) // South
		food[SIDE - 1][i] = -1;
	for (int i = 0; i < SIDE; i++) // East
		food[i][SIDE - 1] = -1;
} // initFood()

void initMoisture(void) {
	// Called by init()
	// Initializes global SIDE-by-SIDE moisture matrix

	for (int i = 0; i < SIDE; i++) {
		// Give each cell the "limited moisture "value
		for (int j = 0; j < SIDE; j++)
			moisture[i][j] = LIMITED;
	}

	for (int i = 0 ; i < MOISTS; i++)
		// Give moist cells the correct moisture value
		moisture[ROWLST_MOISTAREA[i]][COLLST_MOISTAREA[i]] = MOIST_AREA;

	for (int i = 0; i < AWPs; i++) {
		// Alter moisture value near and at AWPs

		int i_row = ROWLST_AWP[i];
		int i_col = COLLST_AWP[i];

		// Initialize cells two cells away to AWP_R2
		int yPositions[16] = {i_col - 2, i_col - 2, i_col - 2, i_col - 2, i_col - 2, i_col - 1, i_col - 1, i_col, i_col, i_col + 1, i_col + 1, i_col + 2, i_col + 2, i_col + 2, i_col + 2, i_col + 2};
		int xPositions[16] = {i_row - 2, i_row - 1, i_row, i_row + 1, i_row + 2, i_row - 2, i_row + 2, i_row - 2, i_row + 2, i_row - 2, i_row + 2, i_row - 2, i_row - 1, i_row, i_row + 1, i_row + 2};
		for (int j = 0; j < 16; j++) {
			for (int k = 0; k < 16; k++) {
				int x_pos = xPositions[j];
				int y_pos = yPositions[k];
				if ( (x_pos >= 0) && (x_pos < SIDE) && (y_pos >= 0) && (y_pos < SIDE) )
					moisture[x_pos][y_pos] = AWP_R2;
			}
		}

		// Initialize adjacent cells to AWP_R1
		int yPositions2[8] = {i_col - 1, i_col - 1, i_col - 1, i_col, i_col, i_col + 1, i_col + 1, i_col + 1};
		int xPositions2[8] = {i_row - 1, i_row, i_row + 1, i_row - 1, i_row + 1, i_row - 1, i_row, i_row + 1};
		for (int j = 0; j < 8; j++) {
			for (int k = 0; k < 8; k++) {
				int x_pos = xPositions2[j];
				int y_pos = yPositions2[k];
				if ( (x_pos >= 0) && (x_pos < SIDE) && (y_pos >= 0) && (y_pos < SIDE) )
					moisture[x_pos][y_pos] = AWP_R1;
			}
		}

		// Initialize the AWP cell's moisture to AWP
		moisture[i_row][i_col] = AWP;
	}

	// Initialize north, south, east borders to moisture value -1
	for (int i = 0; i < SIDE; i++)
		// North border
		moisture[0][i] = -1;
	for (int i = 0; i < SIDE; i++)
		// South border
		moisture[SIDE - 1][i] = -1;
	for (int i = 0; i < SIDE; i++)
		// East border
		moisture[i][SIDE - 1] = -1;
} // initMoisture()

void initToads(void) {
	// Called by init()
	// Initializes global "toads" array with toads at random locations in
	// next-to-last column, excluding the first and last rows.

	for (int i = 0; i < NUM_TOADS; i++) {
		toads[i].toadID = i;
		toads[i].energy = normalDistributionRand(MEAN_ENERGY, STD_ENERGY, 0.0, 1.0);
		toads[i].water =  normalDistributionRand(MEAN_WATER, STD_WATER, 0.0, 1.0);
		toads[i].alive = ALIVE;
		toads[i].row = (rand() % (SIDE-2)) + 1; // int, 1 to SIDE-2
		toads[i].col = SIDE - 2;
		toads[i].rowLast = -1;
		toads[i].colLast = -1;
		toads[i].numTimeSteps = 0;
	}
} // initToads()

void init(void) {
	// Called by CaneToads()
	// Initialize global matrices "food," "moisture," and "toads"

	initFood();
	initMoisture();
	initToads();
} // init()

int getNbrsLst(struct toad aToad, int numNeighbors, int elimPrevious, int* rowLst, int* colLst) {
	// Called by hopForFun(), lookForFoodOrMoisture()
	// Return int n of toad's neighbors, and alter rowLst and colLst such
	// that the first n positions contain the neighbors.

	// Get row and column coordinates of all neighbors
	if (numNeighbors == 4 || numNeighbors == 5) {
		rowLst[0] = aToad.row - 1;
		rowLst[1] = aToad.row;
		rowLst[2] = aToad.row + 1;
		rowLst[3] = aToad.row;
		colLst[0] = aToad.col;
		colLst[1] = aToad.col + 1;
		colLst[2] = aToad.col;
		colLst[3] = aToad.col - 1;
	}

	if (numNeighbors == 5) { // Append this cell's coordinates
		rowLst[4] = aToad.row;
		colLst[4] = aToad.col;
	}

	if (numNeighbors == 8 || numNeighbors == 9) {
		rowLst[0] = aToad.row - 1;
		rowLst[1] = aToad.row - 1;
		rowLst[2] = aToad.row;
		rowLst[3] = aToad.row + 1;
		rowLst[4] = aToad.row + 1;
		rowLst[5] = aToad.row + 1;
		rowLst[6] = aToad.row;
		rowLst[7] = aToad.row + 1;
		colLst[0] = aToad.col;
		colLst[1] = aToad.col - 1;
		colLst[2] = aToad.col - 1;
		colLst[3] = aToad.col - 1;
		colLst[4] = aToad.col;
		colLst[5] = aToad.col + 1;
		colLst[6] = aToad.col + 1;
		colLst[7] = aToad.col + 1;
	}

	if (numNeighbors == 9) { // Append this cell's coordinates
		rowLst[8] = aToad.row;
		colLst[8] = aToad.col;
	}
	
	// Eliminate cells that the toad cannot move to
	// Do this by creating pseudo-copies of rowLst, colLst that are
	// the same except they don't include said cells, then
	// replacing row and col with those pseudo-copies
	int nextPos = 0;
	int size = 0;
	int pseudo_rowLst[numNeighbors];
	int pseudo_colLst[numNeighbors];
	for (int i = 0; i < numNeighbors; i++) {
		int row = rowLst[i];
		int col = colLst[i];
		int toRemove = 0;

		// Check to see if cell is a fenced AWP
		for (int j = 0; j < NUM_FENCED; j++) {
			int AWP_index = FENCED[j];
			int fenced_row = ROWLST_AWP[AWP_index];
			int fenced_col = COLLST_AWP[AWP_index];
			if (fenced_row == row && fenced_col == col)
				toRemove = 1;
		}

		// Check to see if cell is on a boundary
		if (toRemove == 0) {
			if ( (row == 0) || (row == (SIDE - 1)) || (col == 0) || (col == (SIDE - 1)) ) {
				toRemove = 1;
			}
		}

		// If elimPrevious is true, check to see if this
		// cell is the toad's previous location
		if (elimPrevious == 1) {
			if (toRemove == 0) {
				if (row == aToad.rowLast && col == aToad.colLast)
					toRemove = 1;
			}
		}

		// Add cell to pseudo-copy if toad can move there
		if (toRemove == 0) {
			pseudo_rowLst[nextPos] = rowLst[i];
			pseudo_colLst[nextPos] = colLst[i];
			nextPos++;
			size++;
		}
	}

	// If the psuedo-copies have fewer elements than the originals,
	// replace the originals with the pseudo-copies
	if (numNeighbors != size) {
		for (int i = 0; i < size; i++) {
			rowLst[i] = pseudo_rowLst[i];
			colLst[i] = pseudo_colLst[i];
		}
		numNeighbors = size;
	}

	return numNeighbors; // Return the size of the arrays rowLst and colLst
} // getNbrsLst()

void hopToMax(int * rowLst, int * colLst, char gridMat, int size, int * row, int * col) {
	// Called by lookForFoodOrMoisture()
	// gridMat is either 'f' or 'm', meaning food or moisture
	// Change parameters row and col to new toad location with maximum
	// grid value. size is the number of neighbors the cell has.

	// Find grid levels of neighbors
	double valuesAround[size];
	for (int i = 0; i < size; i++) {
		if (gridMat == 'f')
			valuesAround[i] = food[rowLst[i]][colLst[i]];
		else if (gridMat == 'm')
			valuesAround[i] = moisture[rowLst[i]][colLst[i]];
	}
	
	// Find maximum value in valuesAround
	double max = valuesAround[0]; // Maximum value in valuesAround
	for (int i = 1; i < size; i++) {
		if (valuesAround[i] > max)
			max = valuesAround[i];
	}

	int dirLst[size]; // Indices where gridMat[rowLst[i]][colLst[i]] is max
	int numOfIndices = 0; // Number of indices in dirLst which are valid
	for (int i = 0; i < size; i++) {
		if (valuesAround[i] == max) {
			dirLst[numOfIndices] = i;
			numOfIndices++;
		}
	}

	int ndx = rand() % numOfIndices; // If one index, then r = 0
	*row = rowLst[dirLst[ndx]]; // DIRLST
	*col = colLst[dirLst[ndx]];
} // hopToMax()

struct toad useWaterEnergyHopping(struct toad aToad) {
	// Called by lookForFoodOrMoisture() and hopForFun()

	if (moisture[aToad.row][aToad.col] < MOIST_AREA)
		aToad.water = aToad.water - WATER_HOPPING;
	aToad.energy = aToad.energy - ENERGY_HOPPING;
	
	return aToad;
} // useWaterEnergyHopping()

struct toad useWaterEnergySitting(struct toad aToad) {
	// Called by lookForFoodOrMoisture()

	if (moisture[aToad.row][aToad.col] < MOIST_AREA)
		aToad.water = aToad.water - 0.5 * WATER_HOPPING;
	aToad.energy = aToad.energy - 0.5 * ENERGY_HOPPING;

	return aToad;
} // useWaterEnergySitting()

struct toad lookForFoodOrMoisture(struct toad aToad, char foodOrMoisture) {
	// Called by CaneToads(), thirsty(), and veryThirsty()
	// Move toad to neighbor with most food or moisture

	// Get neighbors, including current cell
	int numNeigh;
	int rowLst[NUM_NEIGHBORS + 1];
	int colLst[NUM_NEIGHBORS + 1];
	if (foodOrMoisture == 'f') // food: consider current and previous positions
		numNeigh = getNbrsLst(aToad, NUM_NEIGHBORS + 1, 0, rowLst, colLst);
	else // moisture: don't consider current or previous positions
		numNeigh = getNbrsLst(aToad, NUM_NEIGHBORS, 1, rowLst, colLst);
	int newRowLst[numNeigh];
	int newColLst[numNeigh];
	for (int i = 0; i < numNeigh; i++) {
		newRowLst[i] = rowLst[i];
		newColLst[i] = colLst[i];
	}

	// Store current cell as last cell
	aToad.rowLast = aToad.row;
	aToad.colLast = aToad.col;

	// Hop to "legal" neighboring location with maximum moisture/food
	int row, col;
	hopToMax(newRowLst, newColLst, foodOrMoisture, numNeigh, &row, &col);
	aToad.row = row;
	aToad.col = col;

	// Use energy and perhaps water
	if ((aToad.row == aToad.rowLast) && (aToad.col == aToad.colLast))
		aToad = useWaterEnergySitting(aToad);
	else
		aToad = useWaterEnergyHopping(aToad);

	return aToad;
} // lookForFoodOrMoisture()

struct toad eat(struct toad aToad) {
	// Called by CaneToads()
	// Simulate eating: increase toad's energy and water, return toad

	// Determine the correct amount that the toad should eat
	double amtEat = AMT_EAT;
	if (food[aToad.row][aToad.col] < amtEat)
		amtEat = food[aToad.row][aToad.col];
	if ( (1 - aToad.energy) < amtEat)
		amtEat = 1 - aToad.energy;
	aToad.energy = aToad.energy + amtEat;

	// Determine the correct amount of water that the toad intakes
	double amtDrink = FRACTION_WATER * amtEat;
	if ( (1 - aToad.water) < amtDrink)
		amtDrink = 1 - aToad.water;
	aToad.water = aToad.water + amtDrink;

	// Decrease remaining food in the cell by the amount eaten
	food[aToad.row][aToad.col] = food[aToad.row][aToad.col] - amtEat;

	return aToad;
} // eat()

struct toad drink(struct toad aToad) {
	// Called by CaneToads()
	// Simulate drinking: increase toad's water, return toad

	double amtDrink = AMT_DRINK;
	if ( (1 - aToad.water) < amtDrink)
		amtDrink = 1 - aToad.water;
	aToad.water = aToad.water + amtDrink;

	return aToad;
} // drink()

struct toad veryThirsty(struct toad aToad) {
	// Called by CaneToads()
	// Change the position of a very thirsty toad

	if (moisture[aToad.row][aToad.col] < AWP)
		aToad = lookForFoodOrMoisture(aToad, 'm');

	return aToad;
} // veryThirsty()

struct toad thirsty(struct toad aToad) {
	// Called by CaneToads()
	// Change the state of a thirsty toad

	if (moisture[aToad.row][aToad.col] < MOIST_AREA) // In dry area
		aToad = lookForFoodOrMoisture(aToad, 'm');
	else if (moisture[aToad.row][aToad.col] < AWP) { // In moist area
		aToad.numTimeSteps++;
		double r = (rand() % 100) / 100.0; // random number between 0.0 and 1.0
		if ( 1.0/aToad.numTimeSteps < r) {
			// Leave moist area to look for more moisture
			aToad.numTimeSteps = 0;
			aToad = lookForFoodOrMoisture(aToad, 'm');
		}
	}
	
	return aToad;
} // thirsty()

struct toad hopForFun(struct toad aToad) {
	// Called by CaneToads()
	// Updates a toad's location to simulate hopping randomly

	// Get neighbors, including last cell, excluding current cell
	int rowLst[NUM_NEIGHBORS];
	int colLst[NUM_NEIGHBORS];
	int numNeigh = getNbrsLst(aToad, NUM_NEIGHBORS, 0, rowLst, colLst);
	int newRowLst[numNeigh];
	int newColLst[numNeigh];
	for (int i = 0; i < numNeigh; i++) {
		newRowLst[i] = rowLst[i];
		newColLst[i] = colLst[i];
	}

	// Store current cell as last cell
	//aToad.rowLast = aToad.row;
	//aToad.colLast = aToad.col;

	// Hop to random neighboring location
	int loc = rand() % numNeigh;
	aToad.row = newRowLst[loc];
	aToad.col = newColLst[loc];

	// Reduce energy and water
	aToad = useWaterEnergyHopping(aToad);

	return aToad;
} // hopForFun()

struct toad changeState(struct toad aToad) {
	// Called by CaneToads()
	// Changes state of toad if it starves, desiccates, or migrates

	if (aToad.water < DESICCATE || aToad.energy < STARVE) {
		aToad.alive = DEAD;
		aToad.row = 0;
		aToad.col = 0;
	} else if (aToad.col == 1)
		aToad.alive = MIGRATED;

	return aToad;
} // changeState()

void sendToad(int toadID, double energy, double water, int alive, int row, int col, int rowLast, int colLast, int numTimeSteps, int dest) {
	// Performs multiple MPI_Sends for the toad structure
	MPI_Send(&toadID, 1, MPI_INT, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&energy, 1, MPI_DOUBLE, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&water, 1, MPI_DOUBLE, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&alive, 1, MPI_INT, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&row, 1, MPI_INT, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&col, 1, MPI_INT, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&rowLast, 1, MPI_INT, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&colLast, 1, MPI_INT, dest, TAG, MPI_COMM_WORLD);
	MPI_Send(&numTimeSteps, 1, MPI_INT, dest, TAG, MPI_COMM_WORLD);
}

struct toad receiveToad(int source, MPI_Status status) {
	// Performs multiple MPI_Recvs for the toad structure
	
	struct toad aToad;
	MPI_Recv(&aToad.toadID, 1, MPI_INT, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.energy, 1, MPI_DOUBLE, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.water, 1, MPI_DOUBLE, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.alive, 1, MPI_INT, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.row, 1, MPI_INT, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.col, 1, MPI_INT, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.rowLast, 1, MPI_INT, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.colLast, 1, MPI_INT, source, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&aToad.numTimeSteps, 1, MPI_INT, source, TAG, MPI_COMM_WORLD, &status);
	return aToad;
}

void CaneToads(int my_rank, int world_size, MPI_Status status) {
	// Called by main()
	// Function to perform an entire simulation for one purview in a
	// grid, simulating the toads in this client's purview at each step.
	// Communication with the server happens during the function.

	// Receive my rows
	int numRows, firstRow;
	MPI_Recv(&numRows, 1, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);
	MPI_Recv(&firstRow, 1, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);
	int lastRow = firstRow + numRows - 1;

//	printf("Process %i taking rows %i - %i.\n", my_rank, firstRow, lastRow);

	// Receive and change my food and moisture matrices
	if (firstRow > 0) // Receive row above purview
		MPI_Recv(&moisture[firstRow - 1][0], SIDE, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD, &status);
	for (int i = 0; i < numRows; i++) { // Receive rows in purview
		MPI_Recv(&food[firstRow + i][0], SIDE, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD, &status);
		MPI_Recv(&moisture[firstRow + i][0], SIDE, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD, &status);
	}
	if (lastRow < SIDE - 1) // Receive row below purview
		MPI_Recv(&moisture[lastRow + 1][0], SIDE, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD, &status);

	for (int i = 0; i < NUM_STEPS; i++) { // for each step:
		// Receive boundary rows
		if (firstRow > 0) // Receive row above purview
			MPI_Recv(&food[firstRow - 1][0], SIDE, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD, &status);
		if (lastRow < SIDE - 1) // Receive row below purview
			MPI_Recv(&food[lastRow + 1][0], SIDE, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD, &status);

		// For each toad, receive toad and eat/drink
		int myNumToads;
		MPI_Recv(&myNumToads, 1, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);
		for (int j = 0; j < myNumToads; j++) {
			struct toad thisToad = receiveToad(SERVER, status);
			toads[j] = thisToad;

			if (toads[j].alive == ALIVE) {
				double r1 = (rand() % 100) / 100.0;

				if ( (toads[j].energy < WOULD_LIKE_EAT) || ( (toads[j].water < WOULD_LIKE_DRINK) && (r1 < MAY_EAT)))
					// Toad j would like to eat or is thirsty so may eat
					toads[j] = eat( toads[j] );
				if (moisture[toads[j].row][toads[j].col] >= AWP && toads[i].water < WOULD_LIKE_DRINK)
					// Water is available and toad j would like to drink
					toads[j] = drink( toads[j] );
			}
		}

		// For each toad, move to next position
		for (int j = 0; j < myNumToads; j++) {
			if (toads[j].alive == ALIVE) {
				// Process hopping for water, food, or fun
				double r2 = (rand() % 100) / 100.0;
				if (toads[j].water < NEED_TO_DRINK)
					toads[j] = veryThirsty( toads[j] );
				else if (toads[j].water < WOULD_LIKE_DRINK)
					toads[j] = thirsty( toads[j] );
				else if (toads[j].energy < WOULD_LIKE_EAT)
					toads[j] = lookForFoodOrMoisture(toads[j], 'f');
				else if (r2 < MAY_HOP)
					toads[j] = hopForFun(toads[j]);
				else
					toads[j] = useWaterEnergySitting(toads[j]);
				toads[j] = changeState(toads[j]); // Change state if necessary
			}
		}

		// Send toads back to server
		for (int j = 0; j < myNumToads; j++) {
			sendToad(toads[j].toadID, toads[j].energy, toads[j].water, toads[j].alive, toads[j].row, toads[j].col, toads[j].rowLast, toads[j].colLast, toads[j].numTimeSteps, SERVER);
		}

		// Forward my top and bottom row status to main process
		MPI_Send(&food[firstRow][0], SIDE, MPI_DOUBLE, SERVER, TAG, MPI_COMM_WORLD);
		MPI_Send(&food[lastRow][0], SIDE, MPI_DOUBLE, SERVER, TAG, MPI_COMM_WORLD);

	} // End for each step

} // caneToads()

int main(int argc, char ** argv) {
	// Main function
	// Initialize random number generator

	srand(time(NULL));

	int totalAlive = 0;
	int totalDead = 0;
	int totalMigrated = 0;
	int my_rank, world_size;
	MPI_Status status;

	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &world_size);

	if (my_rank == SERVER)
		printf("Running %i simulations on %i processes.\n", NUM_SIMULATIONS, world_size);

	// For each simulation
	for (int i = 0; i < NUM_SIMULATIONS; i++) {

		if (my_rank == SERVER) {
			int source = SERVER;
			int tag = TAG;

			init(); // Initialize food, moisture, and toads arrays

			// Send setup rows of food and moisture to client
			int row = 0;
			int rowsLeft = SIDE;
			int processesLeft = world_size - 1;
			int rowFirstLast[world_size][2];
			for (int pNum = 1; pNum < world_size; pNum++) {
				int numRows = rowsLeft / processesLeft;
				rowsLeft = rowsLeft - numRows;
				processesLeft = processesLeft - 1;

				int firstRow = row;
				int lastRow = row + numRows - 1;
				rowFirstLast[pNum][0] = firstRow;
				rowFirstLast[pNum][1] = lastRow;

				MPI_Send(&numRows, 1, MPI_INT, pNum, TAG, MPI_COMM_WORLD);
				MPI_Send(&row, 1, MPI_INT, pNum, TAG, MPI_COMM_WORLD);

				if (firstRow > 0) // Send preceeding row's moisture
					MPI_Send(&moisture[firstRow - 1][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD);
				for (int i = firstRow; i <= lastRow; i++) { // Send client's rows
					MPI_Send(&food[i][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD);
					MPI_Send(&moisture[i][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD);
					row++;
				}
				if (lastRow < SIDE - 1) // Send following row's moisture
					MPI_Send(&moisture[lastRow + 1][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD);

			} // End for each process

			for (int step = 0; step < NUM_STEPS; step++) { // For each step
				struct toad newToads[NUM_TOADS];
				int nextPosition = 0;
				for (int pNum = 1; pNum < world_size; pNum++) {
					int firstRow = rowFirstLast[pNum][0];
					int lastRow = rowFirstLast[pNum][1];

//					printf("Process %i: row %i to row %i.\n", pNum, firstRow, lastRow);

					// Send neighboring food rows
					if (firstRow > 0)
						MPI_Send(&food[firstRow - 1][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD);
					if (lastRow < SIDE - 1)
						MPI_Send(&food[lastRow + 1][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD);

					// Send number of toads and send toads
					int toadsFound = 0;
					struct toad toadsToSend[NUM_TOADS];
					for (int i = 0; i < NUM_TOADS; i++) {
						if (toads[i].row >= firstRow && toads[i].row <= lastRow ) {
							toadsToSend[toadsFound] = toads[i];
							toadsFound++;
						}
					}
					MPI_Send(&toadsFound, 1, MPI_INT, pNum, TAG, MPI_COMM_WORLD);
					for (int j = 0; j < toadsFound; j++) {
						sendToad(toadsToSend[j].toadID, toadsToSend[j].energy, toadsToSend[j].water, toadsToSend[j].alive, toadsToSend[j].row, toadsToSend[j].col, toadsToSend[j].rowLast, toadsToSend[j].colLast, toadsToSend[j].numTimeSteps, pNum);
					}
//					printf("Sent %i toads to process %i.\n", toadsFound, pNum);

					// Receive toads
					for (int i = 0; i < toadsFound; i++) {
						struct toad thisToad = receiveToad(pNum, status);
						newToads[nextPosition] = thisToad;
						nextPosition++;
					}

					// Receive top and bottom rows
					MPI_Recv(&food[firstRow][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD, &status);
					MPI_Recv(&food[lastRow][0], SIDE, MPI_DOUBLE, pNum, TAG, MPI_COMM_WORLD, &status);
					
				} // End loop through all clients

				for (int i = 0; i < NUM_TOADS; i++)
					toads[i] = newToads[i];
				

			} // End for each step

			int numAlive = 0;
			int numDead = 0;
			int numMigrated = 0;
			for (int i = 0; i < NUM_TOADS; i++) {
				if (toads[i].alive == ALIVE) {
					numAlive++;
				} else if (toads[i].alive == DEAD) {
					numDead++;
				} else {
					numMigrated++;
				}
			}

			totalAlive = totalAlive + numAlive;
			totalDead = totalDead + numDead;
			totalMigrated = totalMigrated + numMigrated;

		} // END SERVER SECTION
		
		else { // CLIENT SECTION
			CaneToads(my_rank, world_size, status);
		}	

	} // End loop through all simulations

	if (my_rank == SERVER) { // Print out results
		double avgAlive = ((double) totalAlive) / NUM_SIMULATIONS;
		double avgDead = ((double) totalDead) / NUM_SIMULATIONS;
		double avgMigrated = ((double) totalMigrated) / NUM_SIMULATIONS;
		printf("Alive: %f\nDead: %f\nMigrated: %f\n", avgAlive, avgDead, avgMigrated);
	}

	MPI_Finalize();
	return 0;
} // main()
