﻿/* Cane Toads Program, August 2011
 * Whitney Sanders (sanderswe@email.wofford.edu)

 * This version was modified starting 29 September, during thorough checking
 * against the Matlab version. Debugging code was removed to start afresh.

 * DESCRIPTION:
 * To accompany an educational module written by Dr. Angela Shiflet and
 * Dr. George Shiflet, titled "Cane Toads"

 * SPECIFICATIONS:
 * The specifics for the simulation are described in the first group
 * of lines containing values for the amount a toad eats, the probability
 * a toad will hop if it's not thirsty or hungry, and so forth.
 * 
 * OUTPUT:
 * The output is printed to standard output. Output contains the number
 * of toads which are alive, dead, and migrated.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define ALIVE 1 // toad is alive
#define AMT_DRINK 0.05 // max amount toad drinks in 1 time step 0.05
#define AMT_EAT 0.01 // max amount toad eats in 1 time step 0.05
#define AWP 1 // moisture value for water, such as an AWP
#define AWP_R1 0.4 // moisture value of cells neighboring water
#define AWP_R2 0.2 // moisture value of cell 2 cells away from water
#define AWPs 4 // number of AWPs
#define DESICCATE 0.6 // level at which desiccation occurs 0.6
#define DEAD 0 // toad is dead
#define ENERGY_HOPPING 0.002 // max amount of energy used by a hop
#define FOOD_CELL 0.03 // food value for initializing constant food grid
#define LIMITED 0.0 // moisture value for limited amount of moisture
#define MAY_EAT 0.2 // probability of eating if thirsty
#define MAY_HOP 0.5 // probability of hopping if not thirsty/hungry
#define MEAN_ENERGY 0.94 // mean energy for initialized toad
#define MEAN_FOOD 0.1 // mean value for initializing random food grid
#define MEAN_WATER 0.94 // mean water for initialized toad 0.94
#define MIGRATED 2 // toad has migrated off the grid
#define MOIST_AREA 0.5 // moisture value for moist area
#define MOISTS 3 // number of moist areas 3
#define NEED_TO_DRINK 0.8 // water level at which toad needs to drink
#define NUM_FENCED 4 // number of fenced AWPs
#define NUM_TOADS 50 // number of toads 50
#define NUM_NEIGHBORS 8 // 4, or 8 if diagonal moves are allowed
#define NUM_SIMULATIONS 1000
#define NUM_STEPS 200 // number of simulation steps, 300
#define FRACTION_WATER 0.6 // fraction of prey that is water
#define PERMANENT_WATER 1.0 // moisture value for natural water
#define SIDE 20 // number of cells along one side of grid 20
#define STARVE 0.6 // level at which starvation occurs
#define STD_ENERGY 0.02 // standard deviation of toads' initial energy
#define STD_FOOD 0.02 // standard deviation of toad's initial food
#define STD_WATER 0.02 // standard deviation of toad's initial water
#define WATER_HOPPING 0.002 // max amount of water used by a hop
#define WOULD_LIKE_DRINK 0.9 // water level at which toad wants a drink
#define WOULD_LIKE_EAT 0.9 // food level at which toad wants food

// Mathematical constants
#define M_E 2.71828
#define M_PI 3.14159

int ROWLST_AWP[AWPs] = {3,11,15,9}; // array of row indices for AWPs
int COLLST_AWP[AWPs] = {5,13,3,9}; // array of column indices for AWPs
int ROWLST_MOISTAREA[MOISTS] = {13,15,3}; // array of column indices for moist areas
int COLLST_MOISTAREA[MOISTS] = {7,15,17}; // array of row indices for moist areas
int FENCED[NUM_FENCED] = {0,1,2,3}; // array of ROWLST_AWP indices of fenced AWPs

struct toad {
	int toadID; // Identification number
	double energy; // From 0.0 to 1.0 indicating energy level
	double water; // From 0.0 to 1.0 indicating amount of water
	int alive; // Toad's state; 0: dead, 1: alive, 2: migrated
	int row; // Matrix row location of the toad on grid
	int col; // Matrix column location of toad on grid
	int rowLast; // Previous row location, when looking for water
	int colLast; // Previous col location, when looking for water
	int numTimeSteps; // Number of steps in moist area
};

double food[SIDE][SIDE];
double moisture[SIDE][SIDE];
struct toad toads[NUM_TOADS];

/* ---------------------------------------------------
 * Simulation constants and globals have been defined.
 * --------------------------------------------------- */

double normalDistributionRand(double mean, double stdev, double min, double max) {
	// Called by initFood(), initToads()
	// Returns a random number generated on a normal distribution with
	// given mean and standard deviation, within the given interval

	// Algorithm: Box-Muller-Gauss Method
	// a = uniform random number in [0, 2 * pi)
	// b = stdev * sqrt( -2 * ln(c) )
	// c = uniform random number in [0, 1)
	// return r, where:
	// 	r = b * sin(a) + stdev with probability 0.5
	// 	r = b * cos(a) + stdev with probability 0.5

	// 2*pi is approximately 6.283185
	double a = (rand() % 6283185) / 1000000.0;
	double c = (rand() % 1000000) / 1000000.0;
	double b = stdev * sqrt(-2 * log(c));
//	printf("\ta = %.03f, b = %.03f, c = %.03f\n", a, b, c);
	int p = rand() % 2;
	double r;
	if (p == 1) {
		r = ( b * sin(a) ) + mean;
	} else {
		r = ( b * cos(a) ) + mean;
	}
//	printf("Generating %.03f around mean %.03f with stddev %.03f\n", r, mean, stdev);
	if (r > max)
		r = max;
	if (r < min)
		r = min;
	return r;
} // normalDistributionRand()

void initFood(void) {
	// Called by init()
	// Initializes global SIDE-by-SIDE food matrix

	// Give each cell a normally distributed food value, based on
	// MEAN_FOOD and STD_FOOD, from 0.0 to 1.0
	for (int i = 0; i < SIDE; i++) {
		for (int j = 0; j < SIDE; j++) {
			// Option 1: use a normal distribution
			//double val = normalDistributionRand(MEAN_FOOD, STD_FOOD, 0.0, 1.0);
			//food[i][j] = val;

			// Option 2: Use constant food values
			food[i][j] = FOOD_CELL;
		}
	}

	// Initialize north, south, east borders at -1
	for (int i = 0; i < SIDE; i++) // North
		food[0][i] = -1;
	for (int i = 0; i < SIDE; i++) // South
		food[SIDE - 1][i] = -1;
	for (int i = 0; i < SIDE; i++) // East
		food[i][SIDE - 1] = -1;
} // initFood()

void initMoisture(void) {
	// Called by init()
	// Initializes global SIDE-by-SIDE moisture matrix

	for (int i = 0; i < SIDE; i++) {
		// Give each cell the "limited moisture "value
		for (int j = 0; j < SIDE; j++)
			moisture[i][j] = LIMITED;
	}

	for (int i = 0 ; i < MOISTS; i++)
		// Give moist cells the correct moisture value
		moisture[ROWLST_MOISTAREA[i]][COLLST_MOISTAREA[i]] = MOIST_AREA;

	for (int i = 0; i < AWPs; i++) {
		// Alter moisture value near and at AWPs

		int i_row = ROWLST_AWP[i];
		int i_col = COLLST_AWP[i];

		// Initialize cells two cells away to AWP_R2
		int yPositions[16] = {i_col - 2, i_col - 2, i_col - 2, i_col - 2, i_col - 2, i_col - 1, i_col - 1, i_col, i_col, i_col + 1, i_col + 1, i_col + 2, i_col + 2, i_col + 2, i_col + 2, i_col + 2};
		int xPositions[16] = {i_row - 2, i_row - 1, i_row, i_row + 1, i_row + 2, i_row - 2, i_row + 2, i_row - 2, i_row + 2, i_row - 2, i_row + 2, i_row - 2, i_row - 1, i_row, i_row + 1, i_row + 2};
		for (int j = 0; j < 16; j++) {
			for (int k = 0; k < 16; k++) {
				int x_pos = xPositions[j];
				int y_pos = yPositions[k];
				if ( (x_pos >= 0) && (x_pos < SIDE) && (y_pos >= 0) && (y_pos < SIDE) )
					moisture[x_pos][y_pos] = AWP_R2;
			}
		}

		// Initialize adjacent cells to AWP_R1
		int yPositions2[8] = {i_col - 1, i_col - 1, i_col - 1, i_col, i_col, i_col + 1, i_col + 1, i_col + 1};
		int xPositions2[8] = {i_row - 1, i_row, i_row + 1, i_row - 1, i_row + 1, i_row - 1, i_row, i_row + 1};
		for (int j = 0; j < 8; j++) {
			for (int k = 0; k < 8; k++) {
				int x_pos = xPositions2[j];
				int y_pos = yPositions2[k];
				if ( (x_pos >= 0) && (x_pos < SIDE) && (y_pos >= 0) && (y_pos < SIDE) )
					moisture[x_pos][y_pos] = AWP_R1;
			}
		}

		// Initialize the AWP cell's moisture to AWP
		moisture[i_row][i_col] = AWP;
	}

	// Initialize north, south, east borders to moisture value -1
	for (int i = 0; i < SIDE; i++)
		// North border
		moisture[0][i] = -1;
	for (int i = 0; i < SIDE; i++)
		// South border
		moisture[SIDE - 1][i] = -1;
	for (int i = 0; i < SIDE; i++)
		// East border
		moisture[i][SIDE - 1] = -1;
} // initMoisture()

void initToads(void) {
	// Called by init()
	// Initializes global "toads" array with toads at random locations in
	// next-to-last column, excluding the first and last rows.

	for (int i = 0; i < NUM_TOADS; i++) {
		toads[i].toadID = i;
		toads[i].energy = normalDistributionRand(MEAN_ENERGY, STD_ENERGY, 0.0, 1.0);
		toads[i].water =  normalDistributionRand(MEAN_WATER, STD_WATER, 0.0, 1.0);
		toads[i].alive = ALIVE;
		toads[i].row = (rand() % (SIDE-2)) + 1; // int, 1 to SIDE-2
		toads[i].col = SIDE - 2;
		toads[i].rowLast = -1;
		toads[i].colLast = -1;
		toads[i].numTimeSteps = 0;
	}
} // initToads()

void init(void) {
	// Called by CaneToads()
	// Initialize global matrices "food," "moisture," and "toads"

	initFood();
	initMoisture();
	initToads();
} // init()

int getNbrsLst(struct toad aToad, int numNeighbors, int elimPrevious, int* rowLst, int* colLst) {
	// Called by hopForFun(), lookForFoodOrMoisture()
	// Return int n of toad's neighbors, and alter rowLst and colLst such
	// that the first n positions contain the neighbors.

	// Get row and column coordinates of all neighbors
	if (numNeighbors == 4 || numNeighbors == 5) {
		rowLst[0] = aToad.row - 1;
		rowLst[1] = aToad.row;
		rowLst[2] = aToad.row + 1;
		rowLst[3] = aToad.row;
		colLst[0] = aToad.col;
		colLst[1] = aToad.col + 1;
		colLst[2] = aToad.col;
		colLst[3] = aToad.col - 1;
	}

	if (numNeighbors == 5) { // Append this cell's coordinates
		rowLst[4] = aToad.row;
		colLst[4] = aToad.col;
	}

	if (numNeighbors == 8 || numNeighbors == 9) {
		rowLst[0] = aToad.row - 1;
		rowLst[1] = aToad.row - 1;
		rowLst[2] = aToad.row;
		rowLst[3] = aToad.row + 1;
		rowLst[4] = aToad.row + 1;
		rowLst[5] = aToad.row + 1;
		rowLst[6] = aToad.row;
		rowLst[7] = aToad.row + 1;
		colLst[0] = aToad.col;
		colLst[1] = aToad.col - 1;
		colLst[2] = aToad.col - 1;
		colLst[3] = aToad.col - 1;
		colLst[4] = aToad.col;
		colLst[5] = aToad.col + 1;
		colLst[6] = aToad.col + 1;
		colLst[7] = aToad.col + 1;
	}

	if (numNeighbors == 9) { // Append this cell's coordinates
		rowLst[8] = aToad.row;
		colLst[8] = aToad.col;
	}
	
	// Eliminate cells that the toad cannot move to
	// Do this by creating pseudo-copies of rowLst, colLst that are
	// the same except they don't include said cells, then
	// replacing row and col with those pseudo-copies
	int nextPos = 0;
	int size = 0;
	int pseudo_rowLst[numNeighbors];
	int pseudo_colLst[numNeighbors];
	for (int i = 0; i < numNeighbors; i++) {
		int row = rowLst[i];
		int col = colLst[i];
		int toRemove = 0;

		// Check to see if cell is a fenced AWP
		for (int j = 0; j < NUM_FENCED; j++) {
			int AWP_index = FENCED[j];
			int fenced_row = ROWLST_AWP[AWP_index];
			int fenced_col = COLLST_AWP[AWP_index];
			if (fenced_row == row && fenced_col == col)
				toRemove = 1;
		}

		// Check to see if cell is on a boundary
		if (toRemove == 0) {
			if ( (row == 0) || (row == (SIDE - 1)) || (col == 0) || (col == (SIDE - 1)) ) {
				toRemove = 1;
			}
		}

		// If elimPrevious is true, check to see if this
		// cell is the toad's previous location
		if (elimPrevious == 1) {
			if (toRemove == 0) {
				if (row == aToad.rowLast && col == aToad.colLast)
					toRemove = 1;
			}
		}

		// Add cell to pseudo-copy if toad can move there
		if (toRemove == 0) {
			pseudo_rowLst[nextPos] = rowLst[i];
			pseudo_colLst[nextPos] = colLst[i];
			nextPos++;
			size++;
		}
	}

	// If the psuedo-copies have fewer elements than the originals,
	// replace the originals with the pseudo-copies
	if (numNeighbors != size) {
		for (int i = 0; i < size; i++) {
			rowLst[i] = pseudo_rowLst[i];
			colLst[i] = pseudo_colLst[i];
		}
		numNeighbors = size;
	}

	return numNeighbors; // Return the size of the arrays rowLst and colLst
} // getNbrsLst()

void hopToMax(int * rowLst, int * colLst, char gridMat, int size, int * row, int * col) {
	// Called by lookForFoodOrMoisture()
	// gridMat is either 'f' or 'm', meaning food or moisture
	// Change parameters row and col to new toad location with maximum
	// grid value. size is the number of neighbors the cell has.

	// Find grid levels of neighbors
	double valuesAround[size];
	for (int i = 0; i < size; i++) {
		if (gridMat == 'f') {
			valuesAround[i] = food[rowLst[i]][colLst[i]];
		}
		else if (gridMat == 'm') {
			valuesAround[i] = moisture[rowLst[i]][colLst[i]];
		}
	}
	
	// Find maximum value in valuesAround
	double max = valuesAround[0]; // Maximum value in valuesAround
	for (int i = 1; i < size; i++) {
		if (valuesAround[i] > max) {
			max = valuesAround[i];
		}
	}

	int dirLst[size]; // Indices where gridMat[rowLst[i]][colLst[i]] is max
	int numOfIndices = 0; // Number of indices in dirLst which are valid
	for (int i = 0; i < size; i++) {
		if (valuesAround[i] == max) {
			dirLst[numOfIndices] = i;
			numOfIndices++;
		}
	}

	int ndx = rand() % numOfIndices; // If one index, then r = 0
	*row = rowLst[dirLst[ndx]]; // DIRLST
	*col = colLst[dirLst[ndx]];
} // hopToMax()

struct toad useWaterEnergyHopping(struct toad aToad) {
	// Called by lookForFoodOrMoisture() and hopForFun()

	if (moisture[aToad.row][aToad.col] < MOIST_AREA) {
		aToad.water = aToad.water - WATER_HOPPING;
	}
	aToad.energy = aToad.energy - ENERGY_HOPPING;
	
	return aToad;
} // useWaterEnergyHopping()

struct toad useWaterEnergySitting(struct toad aToad) {
	// Called by lookForFoodOrMoisture()

	if (moisture[aToad.row][aToad.col] < MOIST_AREA) {
		aToad.water = aToad.water - 0.5 * WATER_HOPPING;
	}
	aToad.energy = aToad.energy - 0.5 * ENERGY_HOPPING;

	return aToad;
} // useWaterEnergySitting()

struct toad lookForFoodOrMoisture(struct toad aToad, char foodOrMoisture) {
	// Called by CaneToads(), thirsty(), and veryThirsty()
	// Move toad to neighbor with most food or moisture

	// Get neighbors, including current cell
	int numNeigh;
	int rowLst[NUM_NEIGHBORS + 1];
	int colLst[NUM_NEIGHBORS + 1];
	if (foodOrMoisture == 'f') { // food: consider current and previous positions
		numNeigh = getNbrsLst(aToad, NUM_NEIGHBORS + 1, 0, rowLst, colLst);
	}
	else { // moisture: don't consider current or previous positions
		numNeigh = getNbrsLst(aToad, NUM_NEIGHBORS, 1, rowLst, colLst);
	}
	int newRowLst[numNeigh];
	int newColLst[numNeigh];
	for (int i = 0; i < numNeigh; i++) {
		newRowLst[i] = rowLst[i];
		newColLst[i] = colLst[i];
	}

	// Store current cell as last cell
	aToad.rowLast = aToad.row;
	aToad.colLast = aToad.col;

	// Hop to "legal" neighboring location with maximum moisture/food
	int row, col;
	hopToMax(newRowLst, newColLst, foodOrMoisture, numNeigh, &row, &col);
	aToad.row = row;
	aToad.col = col;

	// Use energy and perhaps water
	if ((aToad.row == aToad.rowLast) && (aToad.col == aToad.colLast)) {
		aToad = useWaterEnergySitting(aToad);
	} else {
		aToad = useWaterEnergyHopping(aToad);
	}

	return aToad;
} // lookForFoodOrMoisture()

struct toad eat(struct toad aToad) {
	// Called by CaneToads()
	// Simulate eating: increase toad's energy and water, return toad

	// Determine the correct amount that the toad should eat
	double amtEat = AMT_EAT;
	if (food[aToad.row][aToad.col] < amtEat)
		amtEat = food[aToad.row][aToad.col];
	if ( (1 - aToad.energy) < amtEat)
		amtEat = 1 - aToad.energy;
	aToad.energy = aToad.energy + amtEat;

	// Determine the correct amount of water that the toad intakes
	double amtDrink = FRACTION_WATER * amtEat;
	if ( (1 - aToad.water) < amtDrink)
		amtDrink = 1 - aToad.water;
	aToad.water = aToad.water + amtDrink;

	// Decrease remaining food in the cell by the amount eaten
	food[aToad.row][aToad.col] = food[aToad.row][aToad.col] - amtEat;

	return aToad;
} // eat()

struct toad drink(struct toad aToad) {
	// Called by CaneToads()
	// Simulate drinking: increase toad's water, return toad

	double amtDrink = AMT_DRINK;
	if ( (1 - aToad.water) < amtDrink)
		amtDrink = 1 - aToad.water;
	aToad.water = aToad.water + amtDrink;

	return aToad;
} // drink()

struct toad veryThirsty(struct toad aToad) {
	// Called by CaneToads()
	// Change the position of a very thirsty toad

	if (moisture[aToad.row][aToad.col] < AWP)
		aToad = lookForFoodOrMoisture(aToad, 'm');

	return aToad;
} // veryThirsty()

struct toad thirsty(struct toad aToad) {
	// Called by CaneToads()
	// Change the state of a thirsty toad

	if (moisture[aToad.row][aToad.col] < MOIST_AREA) // In dry area
		aToad = lookForFoodOrMoisture(aToad, 'm');
	else if (moisture[aToad.row][aToad.col] < AWP) { // In moist area
		aToad.numTimeSteps++;
		double r = (rand() % 100) / 100.0; // random number between 0.0 and 1.0
		if ( 1.0/aToad.numTimeSteps < r) {
			// Leave moist area to look for more moisture
			aToad.numTimeSteps = 0;
			aToad = lookForFoodOrMoisture(aToad, 'm');
		}
	}
	
	return aToad;
} // thirsty()

struct toad hopForFun(struct toad aToad) {
	// Called by CaneToads()
	// Updates a toad's location to simulate hopping randomly

	// Get neighbors, including last cell, excluding current cell
	int rowLst[NUM_NEIGHBORS];
	int colLst[NUM_NEIGHBORS];
	int numNeigh = getNbrsLst(aToad, NUM_NEIGHBORS, 0, rowLst, colLst);
	int newRowLst[numNeigh];
	int newColLst[numNeigh];
	for (int i = 0; i < numNeigh; i++) {
		newRowLst[i] = rowLst[i];
		newColLst[i] = colLst[i];
	}

	// Store current cell as last cell
	//aToad.rowLast = aToad.row;
	//aToad.colLast = aToad.col;

	// Hop to random neighboring location
	int loc = rand() % numNeigh;
	aToad.row = newRowLst[loc];
	aToad.col = newColLst[loc];

	// Reduce energy and water
	aToad = useWaterEnergyHopping(aToad);

	return aToad;
} // hopForFun()

struct toad changeState(struct toad aToad) {
	// Called by CaneToads()
	// Changes state of toad if it starves, desiccates, or migrates

	if ( (aToad.water < DESICCATE) || (aToad.energy < STARVE) ) {
		aToad.alive = DEAD;
		aToad.row = 0;
		aToad.col = 0;
	} else if (aToad.col == 1) {
		aToad.alive = MIGRATED;
	}

	return aToad;
} // changeState()

void CaneToads(int nums[3]) {
	// Called by main()
	// Function to perform a cane toad simulation

	init(); // Initialize food, moisture, and toads arrays

	// Do simulation
	for (int step = 0; step < NUM_STEPS; step++) { // For each step

		for (int i = 0; i < NUM_TOADS; i++) { // For each toad
			if (toads[i].alive == ALIVE) {
				double r1 = (rand() % 100) / 100.0;

				if ( (toads[i].energy < WOULD_LIKE_EAT) || ( (toads[i].water < WOULD_LIKE_DRINK) && (r1 < MAY_EAT))) {
					// Toad i would like to eat or is thirsty so may eat
					toads[i] = eat( toads[i] );
				} // end if (toad would like eat) or (toad would like drink and may eat)

				if (moisture[toads[i].row][toads[i].col] >= AWP && toads[i].water < WOULD_LIKE_DRINK) {
					// Water is available and toad i would like to drink
					toads[i] = drink( toads[i] );
				} // end if (water is available) and (toad i would like to drink)
			} // end if toad[i] alive
		} // end for each toad
		
		for (int i = 0; i < NUM_TOADS; i++) {
			if (toads[i].alive == ALIVE) {
				// Process hopping for water, food, or fun
				double r2 = (rand() % 100) / 100.0;
				if (toads[i].water < NEED_TO_DRINK) {
					toads[i] = veryThirsty( toads[i] );
				}
				else if (toads[i].water < WOULD_LIKE_DRINK) {
					toads[i] = thirsty( toads[i] );
				}
				else if (toads[i].energy < WOULD_LIKE_EAT) {
					toads[i] = lookForFoodOrMoisture(toads[i], 'f');
				}
				else if (r2 < MAY_HOP) {
					toads[i] = hopForFun(toads[i]);
				}
				else {
					toads[i] = useWaterEnergySitting(toads[i]);
				}
				toads[i] = changeState(toads[i]); // Change state if necessary
			} // end if toad[i] alive
		} // end for each toad
	}

	int numAlive = 0;
	int numDead = 0;
	int numMigrated = 0;
	for (int i = 0; i < NUM_TOADS; i++) {
		if (toads[i].alive == ALIVE) {
			numAlive++;
		} else if (toads[i].alive == DEAD) {
			numDead++;
		} else {
			numMigrated++;
		}
	}

	nums[0] = numAlive;
	nums[1] = numDead;
	nums[2] = numMigrated;

} // caneToads()

int main(void) {
	// Main function
	// Initialize random number generator

	srand(time(NULL));

	int totalAlive = 0;
	int totalDead = 0;
	int totalMigrated = 0;

	for (int i = 0; i < NUM_SIMULATIONS; i++) {
		int nums[3];
		CaneToads(nums);
		totalAlive = totalAlive + nums[0];
		totalDead = totalDead + nums[1];
		totalMigrated = totalMigrated + nums[2];
	}
	double avgAlive = ((double) totalAlive) / NUM_SIMULATIONS;
	double avgDead = ((double) totalDead) / NUM_SIMULATIONS;
	double avgMigrated = ((double) totalMigrated) / NUM_SIMULATIONS;
	printf("Alive: %f\nDead: %f\nMigrated: %f\n", avgAlive, avgDead, avgMigrated);

	return 0;
} // main()
